
import os
import pickle

import torch
import torch.nn as nn
from lightutils import logger

from .config import DEFAULT_CONFIG, CONFIG


class BasePreprocessor(object):
    def __init__(self):
        pass

    @staticmethod
    def load(path=CONFIG['save_path']):
        preprocessor_path = os.path.join(path, 'preprocessor.pkl')
        with open(preprocessor_path, 'rb') as f:
            preprocessor = pickle.load(f)
        logger.info('loading preprocessor from {}'.format(preprocessor_path))
        preprocessor.save_path = path
        return preprocessor

    def save(self, path=CONFIG['save_path']):
        if not os.path.isdir(path):
            os.mkdir(path)
        config_path = os.path.join(path, 'preprocessor.pkl')
        with open(os.path.join(path, 'preprocessor.pkl'), 'wb') as f:
            pickle.dump(self, f)
        logger.info('saved preprocessor to {}'.format(config_path))


class BaseConfig(object):
    def __init__(self):
        pass

    @staticmethod
    def load(path=DEFAULT_CONFIG['save_path']):
        config_path = os.path.join(path, 'config.pkl')
        with open(config_path, 'rb') as f:
            config = pickle.load(f)
        logger.info('loading config from {}'.format(config_path))
        config.save_path = path
        return config

    def save(self, path=None):
        if not hasattr(self, 'save_path'):
            raise AttributeError('config object must init save_path attr in init method!')
        path = path if path else self.save_path
        if not os.path.isdir(path):
            os.mkdir(path)
        config_path = os.path.join(path, 'config.pkl')
        with open(os.path.join(path, 'config.pkl'), 'wb') as f:
            pickle.dump(self, f)
        logger.info('saved config to {}'.format(config_path))


class BaseModel(nn.Module):
    def __init__(self, args):
        super(BaseModel, self).__init__()
        self.args = args
        self.save_path = args.save_path

    def load(self, path=None):
        path = path if path else self.save_path
        map_location = None if torch.cuda.is_available() else 'cpu'
        model_path = os.path.join(path, 'model.pkl')
        self.load_state_dict(torch.load(model_path, map_location=map_location))
        logger.info('loading model from {}'.format(model_path))

    def save(self, path=None):
        path = path if path else self.save_path
        if not os.path.isdir(path):
            os.mkdir(path)
        model_path = os.path.join(path, 'model.pkl')
        torch.save(self.state_dict(), model_path)
        logger.info('saved model to {}'.format(model_path))




from collections import defaultdict, Counter, OrderedDict
from typing import Union, List, Set

import numpy as np
import torch
import torch.nn as nn

seed = 2020
torch.manual_seed(seed)


def _default_unk_index():
    return 0


class Vocab(object):
    def __init__(self, init_token=None, eos_token=None, pad_token="<pad>", unk_token="<unk>"):
        self.word_count = Counter()
        self.init_token = init_token
        self.eos_token = eos_token
        self.unk_token = unk_token
        self.pad_token = pad_token
        self.itos = list()
        self.stoi = defaultdict()
        self._vectors = None

    def build_vocab(self, min_freq=1, max_size=None, specials_first=True):

        specials = list(OrderedDict.fromkeys(tok for tok in [self.unk_token, self.pad_token, self.init_token,
                                                             self.eos_token] if tok is not None))

        if specials_first:
            self.itos = list(specials)

        words_and_frequencies = sorted(self.word_count.items(), key=lambda tup: tup[0])
        words_and_frequencies.sort(key=lambda tup: tup[1], reverse=True)

        max_size = None if max_size is None else max_size + len(self.itos)
        for word, freq in words_and_frequencies:
            if freq < min_freq or len(self.itos) == max_size:
                break
            self.itos.append(word)

        if not specials_first:
            self.itos.extend(list(specials))

        if '<unk>' in specials:
            self.stoi = defaultdict(_default_unk_index)

        self.stoi.update({tok: i for i, tok in enumerate(self.itos)})

        return self

    def add_words(self, words: Union[List, Set]):
        self.word_count.update(words)
        return self

    def add_word(self, word):
        self.word_count[word] += 1
        return self

    def clear(self):
        self.word_count = Counter()
        self.itos = list()
        self.stoi = defaultdict()

    def __len__(self):
        return len(self.itos)

    def __contains__(self, item):
        return item in self.stoi

    def has_word(self, w):
        return self.__contains__(w)

    def __getitem__(self, w):
        return self.stoi[w]

    def to_index(self, w):
        return self.__getitem__(w)

    def to_word(self, idx):
        if isinstance(idx, int) and 0 <= idx < len(self.itos):
            return self.itos[idx]
        return None

    def __iter__(self):
        for word, idx in self.stoi.items():
            yield word, idx

    def numericalize(self, arr):
        assert len(set(len(row) for row in arr)) == 1, "每一行的长度都要相等！"
        return np.array([[self[x] for x in row] for row in arr])

    def init_vectors(self, dim=3):
        self._vectors = nn.Embedding(len(self), dim)

    def vectorize(self, arr):
        return self._vectors(arr).data

    def reverse(self, arr):
        return [self.to_word(x) for x in arr]




from typing import Union, List
import json

import torch
import pandas as pd
import torch.nn.functional as F
from tqdm import tqdm
from torch.utils.data import DataLoader
from sklearn.metrics import classification_report
from lightutils import logger

from .config import CONFIG
from .model import Config, LinearClassifier
from .metric import get_score
from .data_preprocessor import DataPreprocessor
from .tool import EarlyStopping
from .dataset import ClsDataset

seed = 2020
torch.manual_seed(seed)


class Classifier:
    def __init__(self):
        self._config = None
        self._model = None
        self.preprocessor = None

    def init_preprocessor(self, window_size=0):
        self.preprocessor = DataPreprocessor(window_size=window_size)
        return self.preprocessor

    def train(self, train_set: ClsDataset, dev_set=None, save_path=CONFIG['save_path']):
        class_num = self.preprocessor.class_nums
        train_iter = DataLoader(train_set, batch_size=256)
        item_shape = train_set[0][0].shape
        window_size = item_shape[0]
        feature_dimension = item_shape[1]
        self._config = Config(save_path=save_path, class_num=class_num, window_size=window_size,
                              feature_dimension=feature_dimension)
        linear_classifier = LinearClassifier(self._config)
        self._model = linear_classifier
        opt = torch.optim.Adam(linear_classifier.parameters(), lr=0.01)
        scheduler = torch.optim.lr_scheduler.ExponentialLR(opt, gamma=0.98)
        # scheduler = torch.optim.lr_scheduler.StepLR(opt, 25)

        early_stopping = EarlyStopping(score_mode=dev_set is not None)
        for epoch in range(200):
            self._model.train()
            acc_loss = 0
            for x_i, y_i in tqdm(train_iter):
                opt.zero_grad()
                pred = self._model(x_i)
                item_loss = F.cross_entropy(pred, y_i.reshape((-1)))
                acc_loss += item_loss.item()
                item_loss.backward()
                opt.step()
            scheduler.step()
            logger.info("learning rate is {}".format(opt.param_groups[0]["lr"]))
            logger.info('epoch: {}, acc_loss: {}'.format(epoch, acc_loss))
            if dev_set:
                dev_score = self.score(dev_set)
                logger.info("dev score: {}".format(dev_score))
                if early_stopping(new_record=dev_score):
                    logger.info("reach early stopping patience, break training.")
                    break
            else:
                if early_stopping(new_record=acc_loss):
                    logger.info("reach early stopping patience, break training.")
                    break

        self._config.save()
        self._model.save()
        self.preprocessor.save(path=save_path)

    def score(self, test_set: ClsDataset):
        self._model.eval()
        score_list = []
        test_iter = DataLoader(test_set, batch_size=256)
        for x_i, y_i in tqdm(test_iter):
            item_score = get_score(self._model, x_i, y_i.reshape((-1)))
            score_list.append(item_score)
        res = sum(score_list) / len(score_list)
        return res

    def summary(self, dataset: pd.DataFrame, output_path: str = 'summary.csv'):
        features = dataset[self.preprocessor.feature_columns]
        x = self.preprocessor.transform_feature(features)
        pred = self.predict(x)
        dataset['pred_label'] = [x[0] for x in pred]
        dataset['pred_prob'] = [x[1] for x in pred]
        print(dataset[dataset['label'] == dataset['pred_label']])
        dataset.to_csv(output_path, index=False, encoding='utf8')
        logger.info("预测结果已经写入到{}文件中".format(output_path))

    def report(self, dataset: ClsDataset, output_path: str = None):
        pred = self.predict(dataset.x)
        pred_labels = [x[0] for x in pred]
        true_labels = dataset.y_raw.values.tolist()
        all_labels = self.preprocessor.target_vocab.itos
        if output_path:
            with open(output_path, 'w', encoding='utf8') as file:
                cls_rep = classification_report(true_labels, pred_labels, labels=all_labels, zero_division=1,
                                                output_dict=True)
                json.dump(cls_rep, file, ensure_ascii=False, indent=4)
                logger.info("已经成功将结果写入至{}中".format(output_path))
        return classification_report(true_labels, pred_labels, labels=all_labels, zero_division=1)

    def load(self, save_path=CONFIG['save_path']):
        config = Config.load(save_path)
        model = LinearClassifier(config)
        model.load()
        self._config = config
        self._model = model
        self.preprocessor = DataPreprocessor.load(path=save_path)

    def predict(self, x: Union[List, pd.DataFrame]):
        assert isinstance(self.preprocessor, DataPreprocessor)
        if isinstance(x, list):
            x = self.preprocessor.transform_raw(x)
        elif isinstance(x, pd.DataFrame):
            x = self.preprocessor.transform_feature(x)
        self._model.eval()
        pred = self._model(torch.Tensor(x))
        soft_pred = torch.softmax(pred, dim=1)
        pred_prob, pred_idx = torch.max(soft_pred.cpu().data, dim=1)
        pred_cls = self.preprocessor.reverse_target(pred_idx.tolist())
        pred_prob = map(lambda n: round(n, 4), pred_prob.data.tolist())
        return list(zip(pred_cls, pred_prob))




import torch
from torch.utils.data import Dataset

from .data_preprocessor import DataPreprocessor


class ClsDataset(Dataset):
    def __init__(self, data_preprocessor: DataPreprocessor, x_raw, y_raw):
        self.data_preprocessor = data_preprocessor
        self.x_raw = x_raw
        self.y_raw = y_raw
        self.x = self.data_preprocessor.transform_feature(x_raw)
        self.y = self.data_preprocessor.transform_target(y_raw)

    def __getitem__(self, item):
        return torch.as_tensor(self.x[item]), torch.as_tensor(self.y[item], dtype=torch.long)

    def __len__(self):
        return len(self.x)

    def extend(self, other):
        assert isinstance(other, ClsDataset)
        self.x = torch.cat((self.x, other.x))
        self.y = torch.cat((self.y, other.y))
        return self








import torch.nn as nn

from .base import BaseConfig, BaseModel
from .config import DEVICE, DEFAULT_CONFIG


class Config(BaseConfig):
    def __init__(self, **kwargs):
        super(Config, self).__init__()
        for name, value in DEFAULT_CONFIG.items():
            setattr(self, name, value)
        for name, value in kwargs.items():
            setattr(self, name, value)


class LinearClassifier(BaseModel):
    def __init__(self, args):
        super().__init__(args)

        self.class_num = args.class_num
        self.feature_dimension = args.feature_dimension
        self.window_size = args.window_size
        self.hidden_dimension = 512

        self.seq = nn.Sequential(
            nn.Linear(self.feature_dimension*self.window_size, self.hidden_dimension),
            nn.Dropout(0.5),
            nn.ReLU(),
            # nn.Linear(self.hidden_dimension, self.hidden_dimension),
            # nn.Dropout(0.5),
            # nn.ReLU(),
            # nn.Linear(self.hidden_dimension, self.hidden_dimension),
            # nn.Dropout(0.5),
            # nn.ReLU(),
            nn.Linear(self.hidden_dimension, self.class_num)
        )

    def forward(self, x):
        x = x.reshape((-1, self.window_size*self.feature_dimension))
        x = self.seq(x)
        return x




from typing import Union, List

import torch
import pandas as pd
import numpy as np
from sklearn.preprocessing import StandardScaler
from lightutils import check_file

from .vocab import Vocab
from .base import BasePreprocessor


class NoColumnException(Exception):
    def __init__(self, column_name: str, data_name: str):
        super(NoColumnException, self).__init__()
        self._column_name = column_name
        self._data_name = data_name

    def __str__(self):
        return "no column named {} in {} dataset".format(self._column_name, self._data_name)


class NotFittedException(Exception):
    def __str__(self):
        return "the DataPreprocessor has not been fitted, please fit the data first!"


class DataPreprocessor(BasePreprocessor):
    def __init__(self, window_size=0):
        super(DataPreprocessor, self).__init__()
        self.vocabs = dict()
        self.num_scaler = StandardScaler()
        self.num_label_list = None
        self.target_vocab = None
        self._init_token = '<sos>'
        self._eos_token = '<eos>'
        self._init_num = 0.0
        self._eos_num = 1.0
        self.window_size = window_size
        self.feature_columns: List = []

    def set_target(self, target: Union[List, pd.Series]):
        vocab = Vocab(unk_token=None)
        vocab.add_words(target)
        vocab.build_vocab()
        self.target_vocab = vocab

    def fit(self, dataset: pd.DataFrame):
        if not self.feature_columns:
            self.feature_columns = dataset.columns.tolist()

        # 针对离散类型数据建立词表并使用Embedding转化为稠密向量
        for index in dataset:
            if dataset[index].dtype == 'object' or dataset[index].dtype == 'bool':
                vocab = Vocab(init_token=self._init_token, eos_token=self._eos_token)
                vocab.add_words(dataset[index].tolist())
                vocab.build_vocab()
                vocab.init_vectors()
                self.vocabs[index] = vocab
                print(index)

        # 针对连续性数据进行标准化
        def numeric_filter(dataset):
            return dataset.dtypes[(dataset.dtypes == 'float64') | (dataset.dtypes == 'int64')].index

        numeric_data = dataset[numeric_filter(dataset)]
        self.num_scaler.fit(numeric_data)
        self.num_label_list = numeric_filter(dataset).tolist()

    def fit_from_csv(self, csv_path: str, feature_columns: List, target_column: str):
        check_file(csv_path, 'csv')
        self.feature_columns = feature_columns
        data = pd.read_csv(csv_path)
        feature_data = data[feature_columns]

        self.fit(feature_data)

        if target_column not in data:
            raise NoColumnException(target_column, csv_path)
        else:
            self.set_target(data[target_column].tolist())

    def _check_index(self, indexs):
        for vocab in self.vocabs:
            if vocab not in indexs:
                raise IndexError
        for num_index in self.num_label_list:
            if num_index not in indexs:
                raise IndexError

    def _check_state(self):
        if not self.vocabs:
            raise NotFittedException

    def transform_feature(self, dataset: pd.DataFrame):
        self._check_state()
        self._check_index(list(dataset))
        tensor_list = []
        for index in dataset:
            if index in self.vocabs:
                ios_data = np.array([self.vocabs[index].init_token] * self.window_size).reshape((-1, 1))
                eos_data = np.array([self.vocabs[index].eos_token] * self.window_size).reshape((-1, 1))
                index_data = np.concatenate([ios_data, dataset[index].values.reshape((-1, 1)), eos_data])
                index_num = self.vocabs[index].numericalize(index_data)
                index_vec = self.vocabs[index].vectorize(torch.LongTensor(index_num))
                index_vec = torch.squeeze(index_vec, dim=1)
                tensor_list.append(index_vec)
        stand_num_vec = self.num_scaler.transform(dataset[self.num_label_list])
        num_size = stand_num_vec.shape[-1]
        ios_num = np.array([self._init_num]*self.window_size*num_size).reshape((self.window_size, num_size))
        eos_num = np.array([self._eos_num]*self.window_size*num_size).reshape((self.window_size, num_size))
        stand_num_vec = np.concatenate([ios_num, stand_num_vec, eos_num])
        tensor_list.append(torch.Tensor(stand_num_vec))
        total_data = torch.cat(tensor_list, dim=1).data
        res_data = torch.stack([total_data[i: i+self.window_size*2+1] for i in range(len(total_data) - self.window_size*2)])
        return res_data.data.numpy()

    def transform_target(self, target):
        self._check_state()
        if not isinstance(target, np.ndarray):
            target = np.array(target)
        return self.target_vocab.numericalize(target.reshape((-1, 1)))

    def reverse_target(self, target_nums):
        self._check_state()
        return self.target_vocab.reverse(target_nums)

    def transform_raw(self, raw_data):
        self._check_state()
        df = pd.DataFrame(raw_data, columns=self.feature_columns)
        return self.transform_feature(df)

    @property
    def class_nums(self):
        if self.target_vocab is None:
            return 0
        else:
            return len(self.target_vocab)




import torch
from sklearn.metrics import f1_score, accuracy_score, recall_score, precision_score


def get_score(model, x, y, score_type='f1'):
    metrics_map = {
        'f1': f1_score,
        'p': precision_score,
        'r': recall_score,
        'acc': accuracy_score
    }
    metric_func = metrics_map[score_type] if score_type in metrics_map else metrics_map['f1']
    assert len(x) == len(y)
    vec_predict = model(x)
    soft_predict = torch.softmax(vec_predict, dim=1)
    predict_prob, predict_index = torch.max(soft_predict.cpu().data, dim=1)
    y = y.view(-1).cpu().data.numpy()
    return metric_func(predict_index, y, average='micro')




class EarlyStopping:
    def __init__(self, patience=7, score_mode=True):
        self.patience = patience
        self.no_update_counter = 0
        self.best_record = 10000
        self.score_mode = score_mode
        if self.score_mode:
            self.best_record = -10000

    def __call__(self, new_record):
        self.no_update_counter += 1
        if self.score_mode:
            if new_record > self.best_record:
                self.best_record = new_record
                self.no_update_counter = 0
        else:
            if new_record < self.best_record:
                self.best_record = new_record
                self.no_update_counter = 0

        return self.no_update_counter >= self.patience
